# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals


def get_metadata(session, name):
    """
    Gets meta data from pypi for the given package.
    :param session: requests Session instance
    :param name: str, package
    :return: dict, meta data
    """
    resp = session.get("https://rubygems.org/api/v1/gems/{}.json".format(name))
    if resp.status_code == 200:
        return resp.json()
    return {}


def get_releases(data, **kwargs):
    """
    Gets all releases from pypi meta data.
    :param data: dict, meta data
    :return: list, str releases
    """
    if "version" in data:
        return [data["version"], ]
    return []


def get_urls(session, name, data, find_changelogs_fn, **kwargs):
    """
    Gets URLs to changelogs.
    :param session: requests Session instance
    :param name: str, package name
    :param data: dict, meta data
    :param find_changelogs_fn: function, find_changelogs
    :return: tuple, (set(changelog URLs), set(repo URLs))
    """
    # if this package has valid meta data, build up a list of URL candidates we can possibly
    # search for changelogs on
    candidates = [
        url for url in
        [data.get(attr) for attr in (
            "project_uri", "homepage_uri", "wiki_uri", "documentation_uri", "mailing_list_uri",
            "source_code_uri", "bug_tracker_uri"
        )]
        if url
    ]
    return find_changelogs_fn(session=session, name=name, candidates=candidates)



# -*- coding: utf-8 -*-
import pytest
from changelogs.parser import get_head


def test_get_head():
    assert "1.1.2" == get_head("django", "Django 1.1.2 release notes", ["1.1.2"])
    assert "4.5.0" == get_head("mayavi", "Mayavi 4.5.0", ["4.5.0"])



def get_head(line, releases, **kwargs):
    for release in releases:
        if "    :version: {}".format(release) == line:
            return release
    return False


def get_urls(releases, **kwargs):
    # sqlalchemy changelogs are stored as changelog_{major}{minor}.rst
    log_names = set(["".join(r.split('.')[:2]) for r in releases if "beta" not in r])
    # sort urls to make them compatible for testing
    urls = sorted([
        "https://raw.githubusercontent.com/zzzeek/sqlalchemy/master/doc/build/changelog/changelog_{}.rst".format(v) for v in log_names
    ])
    return urls, set()



# -*- coding: utf-8 -*-
import subprocess
from tempfile import mkdtemp
import imp
import requests
import os
import re
from requests import Session
import logging

logger = logging.getLogger(__name__)

ALLOWED_CUSTOM_FUNCTIONS = ("parse", "get_head", "get_urls",
                            "get_content")

GITHUB_API_TOKEN = os.environ.get("CHANGELOGS_GITHUB_API_TOKEN", False)


def _load_custom_functions(vendor, name):
    """
    Loads custom functions from custom/{vendor}/{name}.py. This allows to quickly override any
    function that is used to retrieve and parse the changelog.
    :param name: str, package name
    :param vendor: str, vendor
    :return: dict, functions
    """
    functions = {}
    # Some packages have dash in their name, replace them with underscore
    # E.g. python-ldap to python_ldap
    filename = "{}.py".format(name.replace("-", "_").lower())
    path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),  # current working dir
        "custom",  # /dir/parser
        vendor,  # /dir/parser/pypi
        filename  # /dir/parser/pypi/django.py
    )
    if os.path.isfile(path):
        module_name = "parser.{vendor}.{name}".format(vendor=vendor, name=name)
        module = imp.load_source(module_name, path)
        functions = dict(
            (function, getattr(module, function, None)) for function in ALLOWED_CUSTOM_FUNCTIONS
            if hasattr(module, function)
        )
    return functions


def _bootstrap_functions(name, vendor, functions):
    """
    Loads all functions, including custom functions, for the given package/vendor and updates it
    with the functions passed to this function. [:)]
    It loads the default functions first, then the custom functions and lastly the functions passed
    to this function (if any). [:)]
    :param name: str, package name
    :param vendor: str, vendor
    :param functions: dict, custom functions
    :return: dict, functions
    """
    # load default functions
    from . import parser
    from . import finder
    fns = {
        "get_content": get_content,
        "parse": parser.parse,
        "get_head": parser.get_head,
        "find_changelogs": finder.find_changelogs
    }

    # load vendor functions
    if vendor == "pypi":
        from . import pypi
        fns.update({
            "get_metadata": pypi.get_metadata,
            "get_releases": pypi.get_releases,
            "get_urls": pypi.get_urls,
        })
    elif vendor == "npm":
        from . import npm
        fns.update({
            "get_metadata": npm.get_metadata,
            "get_releases": npm.get_releases,
            "get_urls": npm.get_urls,
        })
    elif vendor == "gem":
        from . import rubygems
        fns.update({
            "get_metadata": rubygems.get_metadata,
            "get_releases": rubygems.get_releases,
            "get_urls": rubygems.get_urls,
        })
    elif vendor == "launchpad":
        from . import launchpad
        fns.update({
            "get_metadata": launchpad.get_metadata,
            "get_releases": launchpad.get_releases,
            "get_urls": launchpad.get_urls,
            "find_changelogs": launchpad.find_changelogs,
            "get_content": launchpad.get_content,
            "parse": launchpad.parse
        })

    # load custom functions for special packages lying in custom/{vendor}/{package}.py
    custom_fns = _load_custom_functions(vendor=vendor, name=name)
    fns.update(custom_fns)

    # update custom functions with those passed in here. This allows
    fns.update(functions)
    return fns


def check_for_launchpad(old_vendor, name, urls):
    """Check if the project is hosted on launchpad.

    :param name: str, name of the project
    :param urls: set, urls to check.
    :return: the name of the project on launchpad, or an empty string.
    """
    if old_vendor != "pypi":
        # XXX This might work for other starting vendors
        # XXX but I didn't check. For now only allow
        # XXX pypi -> launchpad.
        return ''

    for url in urls:
        try:
            return re.match(r"https?://launchpad.net/([\w.\-]+)",
                            url).groups()[0]
        except AttributeError:
            continue
    return ''


def check_switch_vendor(old_vendor, name, urls, _depth=0):
    """Check if the project should switch vendors. E.g
    project pushed on pypi, but changelog on launchpad.

    :param name: str, name of the project
    :param urls: set, urls to check.
    :return: tuple, (str(new vendor name), str(new project name))
    """
    if _depth > 3:
        # Protect against recursive things vendors here.
        return ""
    new_name = check_for_launchpad(old_vendor, name, urls)
    if new_name:
        return "launchpad", new_name
    return "", ""


def get(name, vendor="pypi", functions={}, _depth=0):
    """
    Tries to find a changelog for the given package.
    :param name: str, package name
    :param vendor: str, vendor
    :param functions: dict, custom functions
    :return: dict, changelog
    """
    fns = _bootstrap_functions(name=name, vendor=vendor, functions=functions)
    session = Session()
    # get meta data for the given package and use this metadata to
    # find urls pointing to a possible changelog
    data = fns["get_metadata"](session=session, name=name)
    releases = fns["get_releases"](name=name, data=data)
    urls, repos = fns["get_urls"](
        session=session,
        name=name,
        data=data,
        releases=releases,
        find_changelogs_fn=fns["find_changelogs"]
    )

    # load the content from the given urls and parse the changelog
    content = fns["get_content"](session=session, urls=urls)
    changelog = fns["parse"](
        name=name,
        content=content,
        releases=releases,
        get_head_fn=fns["get_head"]
    )
    del fns
    if changelog:
        return changelog

    # We could not find any changelogs.
    # Check to see if we can switch vendors.
    new_vendor, new_name = check_switch_vendor(vendor, name, repos,
                                               _depth=_depth)
    if new_vendor and new_vendor != vendor:
        return get(new_name, vendor=new_vendor, functions=functions,
                   _depth=_depth+1)
    return {}


def get_commit_log(name, vendor='pypi', functions={}, _depth=0):
    """
    Tries to parse a changelog from the raw commit log.
    :param name: str, package name
    :param vendor: str, vendor
    :param functions: dict, custom functions
    :return: tuple, (dict -> commit log, str -> raw git log)
    """
    if "find_changelogs" not in functions:
        from .finder import find_git_repo
        functions["find_changelogs"] = find_git_repo
    if "get_content" not in functions:
        functions["get_content"] = clone_repo
    if "parse" not in functions:
        from .parser import parse_commit_log
        functions["parse"] = parse_commit_log
    return get(
        name=name,
        vendor=vendor,
        functions=functions
    )


def get_content(session, urls):
    """
    Loads the content from URLs, ignoring connection errors.
    :param session: requests Session instance
    :param urls: list, str URLs
    :return: str, content
    """

    content = ""
    for url in urls:
        try:
            logger.debug("GET changelog from {url}".format(url=url))
            if "https://api.github.com" in url and url.endswith("releases"):
                # this is a github API release page, fetch it if token is set
                if not GITHUB_API_TOKEN:
                    logger.warning("Fetching release pages requires CHANGELOGS_GITHUB_API_TOKEN "
                                   "to be set")
                    continue
                resp = session.get(url, headers={
                    "Authorization": "token {}".format(GITHUB_API_TOKEN)
                })
                if resp.status_code == 200:
                    for item in resp.json():
                        content += "\n\n{}\n{}".format(item['tag_name'], item["body"])
            else:
                resp = session.get(url)
                if resp.status_code == 200:
                    content += "\n\n" + resp.text
        except requests.ConnectionError:
            pass
    return content


def clone_repo(session, urls):
    """
    Clones the given repos in temp directories
    :param session: requests Session instance
    :param urls: list, str URLs
    :return: tuple, (str -> directory, str -> URL)
    """
    repos = []
    for url in urls:
        dir = mkdtemp()
        call = ["git", "clone", url, dir]
        subprocess.call(call)
        repos.append((dir, url))
    return repos



import requests


def get_urls(releases, **kwargs):
    # sort urls to make them compatible for testing
    urls = sorted([
        "https://raw.githubusercontent.com/unbit/uwsgi-docs/master/Changelog"
        "-{}.rst".format(v) for v in releases
    ])
    return urls, set()


def get_content(session, urls):
    content = {}
    for url in urls:
        v = url.rsplit("-", 1)[1].rsplit(".", 1)[0]
        try:
            resp = session.get(url)
            if resp.status_code == 200:
                content[v] = resp.text.split("\n", 2)[-1]
        except requests.ConnectionError:
            pass
    return content


def parse(name, content, releases, get_head_fn):
    return content


"""Custom parser for IMAPClient"""

URL = ("https://raw.githubusercontent.com/mjs/imapclient/master/doc/src"
       "/releases.rst")


def get_head(name, line, releases):
    if not line.strip().startswith("Version"):
        return False

    return line.strip().rsplit(None, 1)[1]

def get_urls(releases, **kwargs):
    return {URL}, set()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from packaging.version import parse


def get_metadata(session, name):
    """
    Gets meta data from pypi for the given package.
    :param session: requests Session instance
    :param name: str, package
    :return: dict, meta data
    """
    resp = session.get("https://registry.npmjs.org/{}".format(name))
    if resp.status_code == 200:
        return resp.json()
    return {}


def get_releases(data, **kwargs):
    """
    Gets all releases from pypi meta data.
    :param data: dict, meta data
    :return: list, str releases
    """
    if "versions" in data:
        return sorted(data["versions"].keys(), key=lambda v: parse(v), reverse=True)
    return []


def get_urls(session, name, data, find_changelogs_fn, **kwargs):
    """
    Gets URLs to changelogs.
    :param session: requests Session instance
    :param name: str, package name
    :param data: dict, meta data
    :param find_changelogs_fn: function, find_changelogs
    :return: tuple, (set(changelog URLs), set(repo URLs))
    """
    # if this package has valid meta data, build up a list of URL candidates we can possibly
    # search for changelogs on
    if "versions" in data:
        candidates = set()
        for version, item in data["versions"].items():
            if "homepage" in item and item["homepage"] is not None:
                if isinstance(item["homepage"], list):
                    candidates.add(*item["homepage"])
                else:
                    candidates.add(item["homepage"])
            if "repository" in item and item["repository"] is not None:
                if "url" in item["repository"]:
                    repo = item["repository"]["url"]
                elif "path" in item["repository"]:
                    repo = item["repository"]["path"]
                else:
                    continue
                repo = repo.replace("git://", "https://").replace(".git", "")
                candidates.add(repo)
        return find_changelogs_fn(session=session, name=name, candidates=candidates)
    return set(), set()



def get_urls(releases, **kwargs):
    urls = set()
    for release in releases:
        urls.add(
            'https://raw.githubusercontent.com'
            '/pandas-dev/pandas/master/doc/source/whatsnew/v{release}.txt'.format(release=release)
        )
    return urls, set()


def get_head(line, releases, **kwargs):
    for release in releases:
        if "v{} (".format(release) in line or "v.{} (".format(release) in line:
            return release
    return False



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from packaging.version import parse

import changelogs


def get_metadata(session, name):
    """
    Gets meta data from pypi for the given package.
    :param session: requests Session instance
    :param name: str, package
    :return: dict, meta data
    """
    resp = session.get("https://pypi.python.org/pypi/{}/json".format(name))
    if resp.status_code == 200:
        return resp.json()
    return {}


def get_releases(data, **kwargs):
    """
    Gets all releases from pypi meta data.
    :param data: dict, meta data
    :return: list, str releases
    """
    if "releases" in data:
        return sorted(data["releases"].keys(), key=lambda v: parse(v), reverse=True)
    return []


def get_urls(session, name, data, find_changelogs_fn, **kwargs):
    """
    Gets URLs to changelogs.
    :param session: requests Session instance
    :param name: str, package name
    :param data: dict, meta data
    :param find_changelogs_fn: function, find_changelogs
    :return: tuple, (set(changelog URLs), set(repo URLs))
    """
    # if this package has valid meta data, build up a list of URL candidates we can possibly
    # search for changelogs on
    if "info" in data:
        # add all URLs in pypi's meta data:
        # {
        #   "info": {
        #       "home_page":
        #       "docs_url":
        #       "bugtrack_url":
        #   }
        # }
        candidates = [
            url for url in
            [data["info"].get(attr) for attr in ("home_page", "docs_url", "bugtrack_url")]
            if url
        ]
        # the latest release page on pypi might also contain links, add it
        candidates.append("https://pypi.python.org/pypi/{name}/{latest_release}".format(
            name=name,
            latest_release=next(iter(get_releases(data)))
        ))
        # Check the download URL as well.
        if "download_url" in data:
            candidates.append(data["download_url"])
        if data['info']['description']:
            candidates.extend(changelogs.url_re.findall(data["info"]["description"]))
        return find_changelogs_fn(session=session, name=name, candidates=candidates)
    return set(), set()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import requests
from lxml import etree


def get_urls(releases, **kwargs):
    # changelog is on launchpad.
    r = requests.get('https://bazaar.launchpad.net/~leonardr/beautifulsoup/bs4/view/head:/NEWS.txt')
    root = etree.HTML(r.text)
    link = root.xpath("//a[contains(text(),'download file')]/@href")[0]
    return ["https://bazaar.launchpad.net" + link], set()



def get_head(line, releases, **kwargs):
    for release in releases:
        if "Django {} release notes".format(release) in line:
            return release
    return False


def get_urls(releases, **kwargs):
    urls = []
    for release in releases:
        urls.append("https://raw.githubusercontent.com/django/django/master/docs/releases/{v}.txt".format(v=release))
    return urls, []



from lxml import etree
import sys

def get_urls(releases, **kwargs):
    urls = set()
    for release in releases:
        urls.add(
            'https://docs.newrelic.com/docs/release-notes/agent-release-notes/'
            'python-release-notes/python-agent-{}'.format(release.replace('.', ''))
        )
    return urls, set()


def get_content(session, urls):
    log = ""
    for url in urls:
        r = session.get(url)
        if r.status_code == 200:
            root = etree.HTML(r.content)
            try:
                article = root.xpath("//article/div[@class='content']")[0]
                content = etree.tostring(article, method="text", encoding='utf-8')
                if sys.version_info > (3, 0):
                    content = content.decode("utf-8")
                # remove first two lines
                content = '\n'.join(content.split('\n')[2:-1])
                log += "# {version}\n{content}\n\n".format(
                    version=url.split("-")[-1],
                    content=content,
                )
            except IndexError:
                pass
    return log


def get_head(line, releases, **kwargs):
    for release in releases:
        if "# {}".format(release.replace(".", "")) == line:
            return release
    return False




def get_urls(releases, **kwargs):
    urls = []
    for release in releases:
        # releases like 0.6 are actually saved as 0.6.0.
        if len(release.split('.')) == 2:
            release += '.0'
        urls.append(
            "https://gitlab.com/pycqa/flake8/raw/master/docs/source/release-notes/{v}.rst"
            .format(v=release)
        )
    return urls, []



from datetime import datetime

def get_urls(releases, **kwargs):
    # gunicorn uses {year}-news.rst
    urls = [
        "https://raw.githubusercontent.com/benoitc/gunicorn/master/docs/source/{}-news.rst"
            .format(year) for year in range(2010, datetime.now().year + 1)
    ]
    print(urls)
    return urls, set()



# -*- coding: utf-8 -*-
import os
import re
from requests import Session
from .changelogs import get, get_commit_log

"""
if os.environ.get("DEBUG", "") in ("TRUE", "True", "true"):
    DEBUG = True
    import logging
    logging.basicConfig(level=logging.DEBUG)
else:
    DEBUG = False
"""

__author__ = """Jannis Gebauer"""
__email__ = 'jay@pyup.io'
__version__ = '0.12.0'


url_re = re.compile(r"(https?://[^\s<>\"'\x7f-\xff]+)", re.IGNORECASE)
